package mygame;

import com.jme3.app.SimpleApplication;
import com.jme3.bullet.BulletAppState;
import com.jme3.bullet.collision.shapes.SphereCollisionShape;
import com.jme3.bullet.control.CharacterControl;
import com.jme3.input.KeyInput;
import com.jme3.input.controls.ActionListener;
import com.jme3.input.controls.KeyTrigger;
import com.jme3.light.AmbientLight;
import com.jme3.light.DirectionalLight;
import com.jme3.material.Material;
import com.jme3.math.ColorRGBA;
import com.jme3.math.FastMath;
import com.jme3.math.Quaternion;
import com.jme3.math.Vector3f;
import com.jme3.scene.Geometry;
import com.jme3.scene.Mesh;
import com.jme3.scene.Node;
import com.jme3.scene.Spatial;
import com.jme3.scene.shape.Box;
import com.jme3.scene.shape.Sphere;

/**
 * Every JME3 game is an instance of the com.jme3.app.SimpleApplication class.
 * The SimpleApplication class is the simplest example of an application: It
 * manages a 3D scene graph, checks for user input, updates the game state, and
 * automatically draws the scene to the screen. These are the core features of a
 * game engine. You extend this simple application and customize it to create
 * your game.
 *
 * @author G. Amador \& A. Gomes
 */
public class Main extends SimpleApplication
        implements ActionListener {

    private BulletAppState bulletAppState;
    private CharacterControl player;
    private Vector3f walkDirection = new Vector3f();
    private boolean left = false, right = false, up = false, down = false;
    // Temporary vectors used on each frame.
    // They here to avoid instanciating new vectors on each frame
    private Vector3f camDir = new Vector3f();
    private Vector3f camLeft = new Vector3f();
    private Node carNode,
            box1Node, box2Node,
            sphere1Node, sphere2Node;
    private Vector3f max, min;
    private float radius = 1;

    /*
     * You start every JME3 game from the main() method, as every standard Java
     * application: 
     * 1 - Instantiate your SimpleApplication-based class. 
     * 2 - The application's start() method to start the game engine.
     *
     * @param args
     */
    public static void main(String[] args) {
        Main app = new Main();            // instantiate the game
        //app.setDisplayFps(false);       // TODO: UNCOMMENT TO HIDE FPS INFO ON APPLICATION RUN
        //app.setDisplayStatView(false);  // TODO: UNCOMMENT TO HIDE STATS VIEW INFO ON APPLICATION RUN
        //app.setShowSettings(false);     // TODO: UNCOMMENT TO HIDE SETTINGS MENU ON STARTUP
        app.start();                      // start the game
    }

    @Override
    public void simpleInitApp() {
        // We re-use the flyby camera for rotation, while positioning is handled by physics
        viewPort.setBackgroundColor(new ColorRGBA(0.7f, 0.8f, 1f, 1f));
        flyCam.setMoveSpeed(10);
        setUpKeys();
        setUpLight();

        /**
         * Set up Physics
         */
        bulletAppState = new BulletAppState();
        // Since bullet is not (yet) multithreaded or GPU accelerated, the jME3
        // implementation allows to run each physics space on a separate thread
        // that is executed in parallel to rendering.
        //bulletAppState.setThreadingType(BulletAppState.ThreadingType.PARALLEL);   // TODO: UNCOMMENT TO RUN PARALLEL JBULLET 
        stateManager.attach(bulletAppState);
        //bulletAppState.getPhysicsSpace().enableDebug(assetManager);               // TODO: UNCOMMENT TO ENABLE JBULLET DEBUG

        /* create a SkyBox floor of dimension (100, 100, 100) */
        // This quaternion stores a 270 degree rolling rotation around the Y axis
        Quaternion roll270 = new Quaternion();
        roll270.fromAngleAxis(FastMath.PI * 3 / 2, new Vector3f(0, 1, 0));

        // down wall        
        Material downWallMat = new Material(assetManager,
                "Common/MatDefs/Misc/Unshaded.j3md");
        downWallMat.setTexture("ColorMap", assetManager.loadTexture(
                "Textures/Skybox/SkyBox2_down.jpg"));
        Box downWallBox = new Box(100, 0, 100);
        Geometry downWall = new Geometry("downWall", downWallBox);
        downWall.setLocalRotation(roll270);
        downWall.setLocalTranslation(new Vector3f(0, 0, 0));
        downWall.setMaterial(downWallMat);
        rootNode.attachChild(downWall);

        /* adding models to the scene graph */
        Quaternion roll180 = new Quaternion();
        roll180.fromAngleAxis(FastMath.PI, new Vector3f(0, 1, 0));

        // add a red car model
        Spatial carRedGeom = assetManager.loadModel(
                "Models/cars/carRed/Car.obj");
        carNode = new Node("carNode");
        carNode.attachChild(carRedGeom);
        carNode.setLocalRotation(roll180);
        carNode.setLocalTranslation(0, 1, 0.15f);   // translate the model 1 in the Y axis 0.15 in the Z axis 
        carNode.scale(0.35f);                       // reduce the model to near 1/3 of its size
        rootNode.attachChild(carNode);

        /**
         * BEGIN OF EXERCISE 6 (Step 6): complete adding AABB and bounding
         * Spheres in the following TODO's for Step 6.
         */
        max = new Vector3f(1, 1, 1);    // the maximum point that defines the boxes 1 and 2.
        min = new Vector3f(-1, -1, -1); // the minimum point that defines the boxes 1 and 2.

        // AABB 1
        Material box1Mat = new Material(assetManager,
                "Common/MatDefs/Misc/Unshaded.j3md");
        box1Mat.setColor("Color", ColorRGBA.Red);
        box1Mat.getAdditionalRenderState().setWireframe(true);
        //Box box1 = new Box(1, 1, 1);
        Box box1 = new Box(min, max);
        Geometry box1Geom = new Geometry("box1", box1);
        //box1Geom.setLocalTranslation(new Vector3f(0, 1, 0));
        box1Geom.setMaterial(box1Mat);
        box1Node = new Node("box1Node");
        box1Node.attachChild(box1Geom);
        box1Node.setLocalTranslation(new Vector3f(0, 1, 0));
        rootNode.attachChild(box1Node);

        // AABB 2
        // TODO: (Sub-Step 6.1) ADD YOUR CODE HERE!        
        

        // Spheres 1
        Material sphere1Mat = new Material(assetManager,
                "Common/MatDefs/Misc/Unshaded.j3md");
        sphere1Mat.setColor("Color", ColorRGBA.Green);
        sphere1Mat.getAdditionalRenderState().setWireframe(true);
        Sphere sphere1 = new Sphere(20, 20, radius);
        //sphere1.setMode(Mesh.Mode.Lines); // DO NOT UNCOMMENT!!! For spheres does the same as material.getAdditionalRenderState().setWireframe(true);
        Geometry sphere1Geom = new Geometry("sphere1", sphere1);
        sphere1Geom.setMaterial(sphere1Mat);
        sphere1Node = new Node("sphere1Node");
        sphere1Node.attachChild(sphere1Geom);
        sphere1Node.setLocalTranslation(new Vector3f(0, 1, 0));
        rootNode.attachChild(sphere1Node);

        // Spheres 2
        // TODO: (Sub-Step 6.2) ADD YOUR CODE HERE!        
        

        // END OF EXERCISE 6 (Step 6). ----------------------------------------- 

        // We set up collision detection for the player by creating
        // a capsule collision shape and a CharacterControl.
        // The CharacterControl offers extra settings for
        // size, stepheight, jumping, falling, and gravity.
        // We also put the player in its starting position.
        SphereCollisionShape playerCollisionShape = new SphereCollisionShape(2.0f);
        player = new CharacterControl(playerCollisionShape, 0.05f);
        player.setPhysicsLocation(new Vector3f(0, 0, 0));

        // Attach the player to the physics space.
        bulletAppState.getPhysicsSpace().add(player);
    }

    private void setUpLight() {
        // We add light so we see the scene
        AmbientLight al = new AmbientLight();
        al.setColor(ColorRGBA.White.mult(1.3f));
        rootNode.addLight(al);

        DirectionalLight dl = new DirectionalLight();
        dl.setColor(ColorRGBA.White);
        dl.setDirection(new Vector3f(2.8f, -2.8f, -2.8f).normalizeLocal());
        rootNode.addLight(dl);
    }

    /**
     * We over-write some navigational key mappings here, so we can add
     * physics-controlled walking:
     */
    private void setUpKeys() {
        inputManager.addMapping("Left", new KeyTrigger(KeyInput.KEY_A));
        inputManager.addMapping("Right", new KeyTrigger(KeyInput.KEY_D));
        inputManager.addMapping("Up", new KeyTrigger(KeyInput.KEY_W));
        inputManager.addMapping("Down", new KeyTrigger(KeyInput.KEY_S));
        inputManager.addListener(this, "Left");
        inputManager.addListener(this, "Right");
        inputManager.addListener(this, "Up");
        inputManager.addListener(this, "Down");
    }

    /**
     * These are our custom actions triggered by key presses. We do not walk
     * yet, we just keep track of the direction the user pressed.
     */
    @Override
    public void onAction(String binding, boolean isPressed, float tpf) {
        switch (binding) {
            case "Left":
                left = isPressed;
                break;
            case "Right":
                right = isPressed;
                break;
            case "Up":
                up = isPressed;
                break;
            case "Down":
                down = isPressed;
                break;
        }
    }

    /**
     * This is the main event loop--walking happens here. We check in which
     * direction the player is walking by interpreting the camera direction
     * forward (camDir) and to the side (camLeft). The setWalkDirection()
     * command is what lets a physics-controlled player walk. We also make sure
     * here that the camera moves with player.
     */
    @Override
    public void simpleUpdate(float tpf) {
        camDir.set(cam.getDirection()).multLocal(0.6f);
        camLeft.set(cam.getLeft()).multLocal(0.4f);
        walkDirection.set(0, 0, 0);
        if (left) {
            walkDirection.addLocal(camLeft);
        }
        if (right) {
            walkDirection.addLocal(camLeft.negate());
        }
        if (up) {
            walkDirection.addLocal(camDir);
        }
        if (down) {
            walkDirection.addLocal(camDir.negate());
        }
        player.setWalkDirection(walkDirection);

        // Set camera translation to equal player translation
        cam.setLocation(new Vector3f(
                player.getPhysicsLocation().x,
                25, //Set camera Y position
                player.getPhysicsLocation().z));

        // Set car rotation to equal the camera rotation around the Y axis 
        carNode.setLocalRotation(new Quaternion(0, cam.getRotation().getY(), 0, cam.getRotation().getW()));

        // Set car translation to equal player translation
        carNode.setLocalTranslation(
                player.getPhysicsLocation().x,
                carNode.getLocalTranslation().y, //Y position stays the same
                player.getPhysicsLocation().z);

        // AABB with AABB collision test
        /**
         * BEGIN OF EXERCISE 7 (Step 7): uncomment code as indicated in comments
         * in the following TODO's for Step 7.
         */
        // TODO: (Sub-Step 7.1) UNCOMMENT THE FOLLOWING BLOCK WHEN IMPLEMENTED AABB WITH AABB COLLISION TEST         
/*        
        rootNode.detachChild(box1Node);
        rootNode.detachChild(box2Node);
        rootNode.detachChild(sphere1Node);
        rootNode.detachChild(sphere2Node);
        rootNode.attachChild(box1Node);
        rootNode.attachChild(box2Node);

        box1Node.setLocalTranslation(
                player.getPhysicsLocation().x,
                box1Node.getLocalTranslation().y,
                player.getPhysicsLocation().z);

        if (collideAABBwithAABB(box1Node, box2Node)) {
            player.setWalkDirection(walkDirection.negate());
            resetDirections();
        }
*/ 
        // END OF EXERCISE 7 (Step 7). -----------------------------------------        

        // SPHERE with SPHERE collision test
        /**
         * BEGIN OF EXERCISE 8 (Step 8): uncomment code as indicated in comments
         * in the following TODO's for Step 8.
         */
        // TODO: (Sub-Step 8.1) UNCOMMENT THE FOLLOWING BLOCK WHEN IMPLEMENTED SPHERE WITH SPHERE COLLISION TEST IMPLEMENTED        
/*        
        rootNode.detachChild(box1Node);
        rootNode.detachChild(box1Node);
        rootNode.detachChild(box2Node);
        rootNode.detachChild(sphere1Node);
        rootNode.detachChild(sphere2Node);
        rootNode.attachChild(sphere1Node);
        rootNode.attachChild(sphere2Node);

        sphere1Node.setLocalTranslation(
                player.getPhysicsLocation().x,
                sphere1Node.getLocalTranslation().y,
                player.getPhysicsLocation().z);

        if (collideSphereWithSphere(sphere1Node, sphere2Node)) {
            player.setWalkDirection(walkDirection.negate());
            resetDirections();
        }
*/        
        // END OF EXERCISE 8 (Step 8). -----------------------------------------        

        // AABB with SPHERE collision test
        /**
         * BEGIN OF EXERCISE 9 (Step 9): uncomment code as indicated in comments
         * in the following TODO's for Step 9.
         */
        // TODO: (Sub-Step 9.1) UNCOMMENT THE FOLLOWING BLOCK WHEN IMPLEMENTED AABB WITH SPHERE COLLISION TEST IMPLEMENTED                
/*
        rootNode.detachChild(box1Node);
        rootNode.detachChild(box2Node);
        rootNode.detachChild(sphere1Node);
        rootNode.detachChild(sphere2Node);
        rootNode.attachChild(box2Node);
        rootNode.attachChild(sphere1Node);

        sphere1Node.setLocalTranslation(
                player.getPhysicsLocation().x,
                sphere1Node.getLocalTranslation().y,
                player.getPhysicsLocation().z);

        if (collideSphereWithAABB(sphere1Node, box2Node)) {
            player.setWalkDirection(walkDirection.negate());
            resetDirections();
        }
*/       
        // END OF EXERCISE 9 (Step 9). -----------------------------------------        
    }

    private void resetDirections() {
        if (left) {
            left = false;
        }
        if (right) {
            right = false;
        }
        if (up) {
            up = false;
        }
        if (down) {
            down = false;
        }
    }

    private boolean collideAABBwithAABB(Node AABB1, Node AABB2) {

        Vector3f maxb1, maxb2, minb1, minb2;
        maxb1 = max.add(AABB1.getWorldBound().getCenter());
        minb1 = min.add(AABB1.getWorldBound().getCenter());
        maxb2 = max.add(AABB2.getWorldBound().getCenter());
        minb2 = min.add(AABB2.getWorldBound().getCenter());

        // AABB with AABB collision test.
        // TODO: (Sub-Step 7.2) ADD YOUR CODE HERE!        
        return false;
    }

    private boolean collideSphereWithSphere(Node Sphere1, Node Sphere2) {
        Vector3f barycentreSphere1 = Sphere1.getWorldBound().getCenter();
        Vector3f barycentreSphere2 = Sphere2.getWorldBound().getCenter();

        // SPHERE with SPHERE collision test.
        // TODO: (Sub-Step 8.2) ADD YOUR CODE HERE!        
        return false;
    }

    private boolean collideSphereWithAABB(Node Sphere, Node AABB) {
        Vector3f barycentreSphere = Sphere.getWorldBound().getCenter();
        Vector3f barycentreAABB = AABB.getWorldBound().getCenter();
        Vector3f AABBmax = max.add(barycentreAABB);
        Vector3f AABBmin = min.add(barycentreAABB);
        float halfDistanceAABB = AABBmax.distance(barycentreAABB);

        Vector3f v1 = new Vector3f(AABBmin);
        Vector3f v2 = new Vector3f(AABBmin.getX(), AABBmax.getY(), AABBmin.getZ());
        Vector3f v3 = new Vector3f(AABBmin.getX(), AABBmin.getY(), AABBmax.getZ());
        Vector3f v4 = new Vector3f(AABBmin.getX(), AABBmax.getY(), AABBmax.getZ());
        Vector3f v5 = new Vector3f(AABBmax.getX(), AABBmin.getY(), AABBmin.getZ());
        Vector3f v6 = new Vector3f(AABBmax.getX(), AABBmin.getY(), AABBmax.getZ());
        Vector3f v7 = new Vector3f(AABBmax.getX(), AABBmax.getY(), AABBmin.getZ());
        Vector3f v8 = new Vector3f(AABBmax);

        float distance0 = Float.POSITIVE_INFINITY;
        float distance1 = barycentreSphere.distance(v1);
        float distance2 = barycentreSphere.distance(v2);
        float distance3 = barycentreSphere.distance(v3);
        float distance4 = barycentreSphere.distance(v4);
        float distance5 = barycentreSphere.distance(v5);
        float distance6 = barycentreSphere.distance(v6);
        float distance7 = barycentreSphere.distance(v7);
        float distance8 = barycentreSphere.distance(v8);

        distance0 = distance1 > distance0 ? distance0 : distance1;
        distance0 = distance2 > distance0 ? distance0 : distance2;
        distance0 = distance3 > distance0 ? distance0 : distance3;
        distance0 = distance4 > distance0 ? distance0 : distance4;
        distance0 = distance5 > distance0 ? distance0 : distance5;
        distance0 = distance6 > distance0 ? distance0 : distance6;
        distance0 = distance7 > distance0 ? distance0 : distance7;
        distance0 = distance8 > distance0 ? distance0 : distance8;

        // AABB with SPHERE collision test.        
        // TODO: (Sub-Step 9.2) ADD YOUR CODE HERE!        
        return false;
    }
}
